/* exec_prog.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>

#include "pcode.h"
#include "vec/vec.h"

static int check_range(data_t* val)
{
    /* Does this machine wrap numbers or die on overflow? */
    return mpz_cmp(val->val, data_min.val) >= 0 &&
           mpz_cmp(val->val, data_max.val) <= 0;
}

static opcode_ret_t opcode_move(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_set(vm->data[insn->arg[2]].val, vm->data[insn->arg[0]].val);
    return RET_NEXT;
}

static opcode_ret_t opcode_add(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_add(vm->data[insn->arg[2]].val,
            vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val);
    if (!check_range(&vm->data[insn->arg[2]])) {
        fprintf(stderr, "Addition result out of bounds\n");
        return RET_FAIL;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_sub(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_sub(vm->data[insn->arg[2]].val,
            vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val);
    if (!check_range(&vm->data[insn->arg[2]])) {
        fprintf(stderr, "Subtraction result out of bounds\n");
        return RET_FAIL;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_mul(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_mul(vm->data[insn->arg[2]].val,
            vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val);
    if (!check_range(&vm->data[insn->arg[2]])) {
        fprintf(stderr, "Multiplication result out of bounds\n");
        return RET_FAIL;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_div(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp_si(vm->data[insn->arg[1]].val, 0) == 0) {
        fprintf(stderr, "Division by zero\n");
        return RET_FAIL;
    } else {
        mpz_tdiv_q(vm->data[insn->arg[2]].val,
                vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val);
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_sqr(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_mul(vm->data[insn->arg[2]].val,
            vm->data[insn->arg[0]].val, vm->data[insn->arg[0]].val);
    if (!check_range(&vm->data[insn->arg[2]])) {
        fprintf(stderr, "Square result out of bounds\n");
        return RET_FAIL;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_sqrt(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp_si(vm->data[insn->arg[1]].val, 0) < 0) {
        fprintf(stderr, "Square root of negative number\n");
        return RET_FAIL;
    } else {
        mpz_sqrt(vm->data[insn->arg[2]].val, vm->data[insn->arg[0]].val);
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_jeq(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp(vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val) == 0) {
        vm->pc = insn->arg[2];
        return RET_GOTO;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_jne(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp(vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val) != 0) {
        vm->pc = insn->arg[2];
        return RET_GOTO;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_jge(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp(vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val) >= 0) {
        vm->pc = insn->arg[2];
        return RET_GOTO;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_jlt(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    if (mpz_cmp(vm->data[insn->arg[0]].val, vm->data[insn->arg[1]].val) < 0) {
        vm->pc = insn->arg[2];
        return RET_GOTO;
    }
    return RET_NEXT;
}

static opcode_ret_t opcode_load(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_t addr;

    mpz_init(addr);
    mpz_add_ui(addr, vm->data[insn->arg[1]].val, insn->arg[0]);

    if (mpz_cmp_si(addr, 0) < 0 || mpz_cmp_si(addr, 999) > 0) {
        fprintf(stderr, "Load from out-of-bounds memory address\n");
        mpz_clear(addr);
        return RET_FAIL;
    } else {
        mpz_set(vm->data[insn->arg[2]].val, vm->data[mpz_get_si(addr)].val);
        mpz_clear(addr);
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_store(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    mpz_t addr;

    mpz_init(addr);
    mpz_add_ui(addr, vm->data[insn->arg[2]].val, insn->arg[1]);

    if (mpz_cmp_si(addr, 0) < 0 || mpz_cmp_si(addr, 999) > 0) {
        fprintf(stderr, "Store to out-of-bounds memory address\n");
        mpz_clear(addr);
        return RET_FAIL;
    } else {
        mpz_set(vm->data[mpz_get_si(addr)].val, vm->data[insn->arg[0]].val);
        mpz_clear(addr);
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_loop(insn_t* insn, interp_t* interp)
{
    /* This function behaves differently than its analogue in Szczurowska,
     * which tests then increments.  Here I increment then test, as is
     * described in MacLennan. */
    vm_t* vm = &interp->vm;

    mpz_add_ui(vm->data[insn->arg[0]].val, vm->data[insn->arg[0]].val, 1);
    if (!check_range(&vm->data[insn->arg[0]])) {
        fprintf(stderr, "Loop increment result out of bounds\n");
        return RET_FAIL;
    } else if (mpz_cmp(vm->data[insn->arg[0]].val,
                vm->data[insn->arg[1]].val) < 0) {
        vm->pc = insn->arg[2];
        return RET_GOTO;
    } else {
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_read(insn_t* insn, interp_t* interp)
{
    vm_t* vm = &interp->vm;
    data_t* card;

    card = VEC_POP1(data_t, &vm->cards);
    if (card == NULL) {
        fprintf(stderr, "No more input cards to read\n");
        return RET_FAIL;
    } else {
        mpz_set(vm->data[insn->arg[2]].val, card->val);
        mpz_clear(card->val);
        return RET_NEXT;
    }
}

static opcode_ret_t opcode_print(insn_t* insn, interp_t* interp)
{
    gmp_fprintf(interp->output, "%+011Zd\n", interp->vm.data[insn->arg[0]].val);
    return RET_NEXT;
}

static opcode_ret_t opcode_halt(insn_t* insn, interp_t* interp)
{
    return RET_HALT;
}

static opcode_ret_t opcode_nval(insn_t* insn, interp_t* interp)
{
    fprintf(stderr, "Undefined opcode executed\n");
    return RET_FAIL;
}

opcode_t opcodes[20] = {
    {"+0", "move" , 1, opcode_move , {ARG_DATA, ARG_IGN , ARG_DATA}},
    {"-0", "nval1", 0, opcode_nval , {ARG_IGN , ARG_IGN , ARG_IGN }},
    {"+1", "add"  , 1, opcode_add  , {ARG_DATA, ARG_DATA, ARG_DATA}},
    {"-1", "sub"  , 1, opcode_sub  , {ARG_DATA, ARG_DATA, ARG_DATA}},
    {"+2", "mul"  , 1, opcode_mul  , {ARG_DATA, ARG_DATA, ARG_DATA}},
    {"-2", "div"  , 1, opcode_div  , {ARG_DATA, ARG_DATA, ARG_DATA}},
    {"+3", "sqr"  , 1, opcode_sqr  , {ARG_DATA, ARG_IGN , ARG_DATA}},
    {"-3", "sqrt" , 1, opcode_sqrt , {ARG_DATA, ARG_IGN , ARG_DATA}},
    {"+4", "jeq"  , 1, opcode_jeq  , {ARG_DATA, ARG_DATA, ARG_TEXT}},
    {"-4", "jne"  , 1, opcode_jne  , {ARG_DATA, ARG_DATA, ARG_TEXT}},
    {"+5", "jge"  , 1, opcode_jge  , {ARG_DATA, ARG_DATA, ARG_TEXT}},
    {"-5", "jlt"  , 1, opcode_jlt  , {ARG_DATA, ARG_DATA, ARG_TEXT}},
    {"+6", "load" , 1, opcode_load , {ARG_BASE, ARG_IDX , ARG_DATA}},
    {"-6", "store", 1, opcode_store, {ARG_DATA, ARG_BASE, ARG_IDX }},
    {"+7", "loop" , 1, opcode_loop , {ARG_DATA, ARG_DATA, ARG_TEXT}},
    {"-7", "nval2", 0, opcode_nval , {ARG_IGN , ARG_IGN , ARG_IGN }},
    {"+8", "read" , 1, opcode_read , {ARG_IGN , ARG_IGN , ARG_DATA}},
    {"-8", "print", 1, opcode_print, {ARG_DATA, ARG_IGN , ARG_IGN }},
    {"+9", "halt" , 1, opcode_halt , {ARG_IGN , ARG_IGN , ARG_IGN }},
    {"-9", "nval3", 0, opcode_nval , {ARG_IGN , ARG_IGN , ARG_IGN }}
};

int exec_prog(interp_t* interp)
{
    vm_t* vm = &interp->vm;
    insn_t* insn;

    vm->pc = 0;

    while (1) {
        if (interp->debug && debug_step(interp) == -1) return -1;

        insn = &vm->text[vm->pc];
        if (insn->opcode == NULL) {
            fprintf(stderr, "Execution of uninitialized program text\n");
            return -1;
        } else {
            switch (insn->opcode->func(insn, interp)) {
            case RET_NEXT: ++vm->pc; if (vm->pc == 1000) vm->pc = 0;
            case RET_GOTO: break;
            case RET_HALT: return  0;
            case RET_FAIL: return -1;
            }
        }
    }
}

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
